//  Copyright 2019 Kakao Corp.
//
//  Licensed under the Apache License, Version 2.0 (the "License");
//  you may not use this file except in compliance with the License.
//  You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
//  Unless required by applicable law or agreed to in writing, software
//  distributed under the License is distributed on an "AS IS" BASIS,
//  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//  See the License for the specific language governing permissions and
//  limitations under the License.

import UIKit

import SafariServices

import RxDataSources

import RxSwift
import RxCocoa
import Alamofire
import RxAlamofire

//import RxKakaoPartnerSDK


import KakaoSDKCommon
import RxKakaoSDKCommon
import KakaoSDKAuth
import RxKakaoSDKAuth
import KakaoSDKUser
import RxKakaoSDKUser
import KakaoSDKTalk
import RxKakaoSDKTalk
import RxKakaoSDKStory
import RxKakaoSDKLink
import KakaoSDKNavi
import KakaoSDKTemplate

import KakaoPartnerSDKCommon
import RxKakaoPartnerSDKUser
import KakaoPartnerSDKTalk
import RxKakaoPartnerSDKTalk
import KakaoPartnerSDKAuth
import RxKakaoPartnerSDKAuth

import KakaoPartnerSDKFriend
import RxKakaoPartnerSDKFriend

extension RxPartnerSDKSampleViewController {
    func updateResultIndicator(cell:UITableViewCell?, row:Row) {
        
        switch(row.apiResult) {
        case .Ready: cell?.backgroundColor = UIColor.white
        case .Success: cell?.backgroundColor = UIColor.green
        case .Fail: cell?.backgroundColor = UIColor.red
        }
    }
    
    func updateResultIndicator() {
        if let resultIndicator = self.resultIndicator {
            let cell = self.tableView.cellForRow(at: resultIndicator.0)
            self.updateResultIndicator(cell: cell, row: resultIndicator.1)
        }
    }
    
    
    func success(_ object: Any? = nil, updateIndicator: Bool = true) {
        print("\(String(describing:object))")
        
        if (updateIndicator) {
            self.resultIndicator?.1.apiResult = .Success
            self.updateResultIndicator()
        }
    }
    
    
    func errorHandler(error:Error?, updateIndicator: Bool = true) {
        guard error != nil else { return }
        
        if (updateIndicator) {
            self.resultIndicator?.1.apiResult = .Fail
            self.updateResultIndicator()
        }
        
        if let sdkError = error as? SdkError {
            print("\n\n[sdk error handling start]------------------------------------------------------------------------------")
            
            switch(sdkError) {
            case .ApiFailed:
                let apiError = sdkError.getApiError()
                switch(apiError.reason) {
                default:
                    print("api error: \(String(describing: error))")
                }
                break
            case .AuthFailed:
                let authError = sdkError.getAuthError()
                switch(authError.reason) {
                default:
                    print("auth error: \(String(describing: error))")
                }
                break
            case .ClientFailed:
                print("client error: \(String(describing: error))")
                break
            }
            print("[sdk error handling end]--------------------------------------------------------------------------------\n")
        }
        else {
            print("\n\n[not sdk error start]------------------------------------------------------------------------------")
            if let nsError = error as NSError? {
                print("normal error: \(String(describing: nsError))")
            }
            print("[not sdk error end]------------------------------------------------------------------------------\n")
        }
    }
}

public class RxPartnerSDKSampleViewController: SDKSampleBaseViewController {
    
    let disposeBag = DisposeBag()
    
    lazy public var dataSource = RxTableViewSectionedReloadDataSource<SectionOfRow>(
        configureCell: { (dataSource, tableView, indexPath, item) in
            let cell = tableView.dequeueReusableCell(withIdentifier: "Cell", for: indexPath)
            
            self.updateResultIndicator(cell:cell, row:item)
            
//            cell.selectionStyle = UITableViewCell.SelectionStyle.none
            cell.textLabel?.adjustsFontSizeToFitWidth = true
            cell.textLabel?.text = "\(item.title)"
            cell.textLabel?.backgroundColor = UIColor.init(white: 1.0, alpha: 0)
            cell.textLabel?.textColor = UIColor.black
            return cell
        })
    
    var resultIndicator : (IndexPath, Row)?
    
    override public func loadView() {
        super.loadView()
        
        self.safeArea = view.layoutMarginsGuide
        setupViews()
        self.tableView.register(UITableViewCell.self, forCellReuseIdentifier: "Cell")
    }
    
    override public func viewDidLoad() {
        super.viewDidLoad()
                
        self.dataSource.titleForHeaderInSection = { dataSource, index in
            return dataSource.sectionModels[index].header
        }
        
        setMenu()
        
        bindViews()
        
        subscribe2()
    }
    
    private func bindViews() {
        Observable.just(self.sections!)
            .bind(to: tableView.rx.items(dataSource: dataSource))
            .disposed(by: disposeBag)
    }
    
//    func subscribe() {
//        self.tableView.rx.modelSelected(Row.self).subscribe(onNext: { row in
//            if let action = row.action {
//                action()
//            }
//        })
//        .disposed(by: disposeBag)
//    }

    func subscribe2() {
        Observable
            .zip(self.tableView.rx.itemSelected, self.tableView.rx.modelSelected(Row.self),
                 resultSelector: { indexPath, row in
                    if let action = row.action {
                        action()
                        
                        self.resultIndicator = (indexPath, row)
                        
                        DispatchQueue.main.asyncAfter(deadline: .now() + 0.1 * 1) {
                            self.tableView.deselectRow(at: indexPath, animated: false)
                        }
                    }
            })
            .subscribe()
            .disposed(by: self.disposeBag)
        
        
    }
    
    public func setTitle(title:String = "Partner SDK Sample", args: String? = nil) {
        self.navigationItem.title = "\(title)(\(args ?? ""))"
    }
    
    func extractPickerElements(friends:Friends<PartnerFriend>) -> [(URL?, String?, Any)]? {
        return friends.elements?.map({ (friend) -> (URL?, String?, Any) in
            return (friend.profileThumbnailImage, friend.profileNickname, friend.uuid)
        })
    }
    
    func extractPickerElements(chats:Chats) -> [(URL?, String?, Any)]? {
        return chats.elements?.map({ (chat) -> (URL?, String?, Any) in
            
            var chatImageUrl : URL? = nil
            if let imageUrl = chat.imageUrl {
                chatImageUrl = imageUrl
            }
            
            if let displayMemberImageUrl = chat.displayMemberImages?[0] {
                chatImageUrl = displayMemberImageUrl
            }
            
            return (chatImageUrl, chat.title, chat.id)
        })
    }
    
    public func getMostTopViewController(base: UIViewController? = UIApplication.shared.keyWindow?.rootViewController) -> UIViewController? {

        if let naviController = base as? UINavigationController {
            return getMostTopViewController(base: naviController.visibleViewController)

        } else if let tabbarController = base as? UITabBarController, let selected = tabbarController.selectedViewController {
            return getMostTopViewController(base: selected)

        } else if let presented = base?.presentedViewController {
            return getMostTopViewController(base: presented)
        }
        return base
    }
    
    @objc
    func close() {
    }
    
    func setMenu() {
        self.sections = [
            // MARK: - section FRIEND start =============================================================================================
            SectionOfRow(header: "FRIEND", items: [
                Row(title: "selectChat(params:) - .chat", action: {
                    let pickerChatRequestParams = PickerChatRequestParams(selectionType:.chat,                    //required  .chat or .chatMember
                                                                          title: nil,                             //default .chat:"채팅방 선택",  .chatMember:"친구 선택"
                                                                          pickerChatFilters: [.regular, .open],   //default [.regular]
                                                                          viewAppearance: .auto,                  //default .auto
                                                                          orientation: .auto,                     //default .auto
                                                                          enableSearch: true)                     //default true
                    
                    PickerApi.shared.rx.selectChat(params:pickerChatRequestParams)
                        .subscribe (onNext:{ (selectedInfo) in
                            self.success(selectedInfo)
                            
                            let (_, selectedChats) = selectedInfo
                            _ = selectedChats

                        }, onError: {error in
                            self.errorHandler(error: error)
                        })
                        .disposed(by: self.disposeBag)
                }),
                Row(title: "selectChatPopup(params:) - .chat", action: {
                    let pickerChatRequestParams = PickerChatRequestParams(selectionType:.chat,                    //required  .chat or .chatMember
                                                                          title: nil,                             //default .chat:"채팅방 선택",  .chatMember:"친구 선택"
                                                                          pickerChatFilters: [.regular, .open],   //default [.regular]
                                                                          viewAppearance: .auto,                  //default .auto
                                                                          orientation: .auto,                     //default .auto
                                                                          enableSearch: true)                     //default true
                    
                    PickerApi.shared.rx.selectChatPopup(params:pickerChatRequestParams)
                        .subscribe (onNext:{ (selectedInfo) in
                            self.success(selectedInfo)
                            
                            let (_, selectedChats) = selectedInfo
                            _ = selectedChats

                        }, onError: {error in
                            self.errorHandler(error: error)
                        })
                        .disposed(by: self.disposeBag)
                }),
                Row(title: "selectChat(params:) - .chatMember", action: {
                    let pickerChatRequestParams = PickerChatRequestParams(selectionType:.chatMember,              //required  .chat or .chatMember
                                                                          title: nil,                             //default .chat:"채팅방 선택",  .chatMember:"친구 선택"
                                                                          pickerChatFilters: [.regular],          //default [.regular],   can't use [.open]
                                                                          viewAppearance: .auto,                  //default .auto
                                                                          orientation: .auto,                     //default .auto
                                                                          enableSearch: true,                     //default true
                                                                          showFriendsOnly: nil,                   //default false
                                                                          disableSelectReasons: [.msgBlocked],    //defalut nil
                                                                          displayAllProfile: nil,                 //default false
                                                                          maxPickableCount: nil,                  //default 30
                                                                          minPickableCount: nil)                  //default 1
                    
                    PickerApi.shared.rx.selectChat(params:pickerChatRequestParams)
                        .subscribe (onNext:{ (selectedInfo) in
                            self.success(selectedInfo)
                            
                            let (selectedUsers, _) = selectedInfo
                            _ = selectedUsers

                        }, onError: {error in
                            self.errorHandler(error: error)
                        })
                        .disposed(by: self.disposeBag)
                }),
                Row(title: "selectChatPopup(params:) - .chatMember", action: {
                    let pickerChatRequestParams = PickerChatRequestParams(selectionType:.chatMember,              //required  .chat or .chatMember
                                                                          title: nil,                             //default .chat:"채팅방 선택",  .chatMember:"친구 선택"
                                                                          pickerChatFilters: [.regular],          //default [.regular],   can't use [.open]
                                                                          viewAppearance: .auto,                  //default .auto
                                                                          orientation: .auto,                     //default .auto
                                                                          enableSearch: true,                     //default true
                                                                          showFriendsOnly: nil,                   //default false
                                                                          disableSelectReasons: [.msgBlocked],    //defalut nil
                                                                          displayAllProfile: nil,                 //default false
                                                                          maxPickableCount: nil,                  //default 30
                                                                          minPickableCount: nil)                  //default 1
                    
                    PickerApi.shared.rx.selectChatPopup(params:pickerChatRequestParams)
                        .subscribe (onNext:{ (selectedInfo) in
                            self.success(selectedInfo)
                            
                            let (selectedUsers, _) = selectedInfo
                            _ = selectedUsers                            

                        }, onError: {error in
                            self.errorHandler(error: error)
                        })
                        .disposed(by: self.disposeBag)
                }),
                Row(title: "selectFriends(params:)", action: {
                    let pickerFriendRequestParams = PickerFriendRequestParams(title: "샘플앱 친구피커",                     //default "카카오톡 친구 선택"
                                                                    pickerFriendOrder: nil,                 //default nil
                                                                    pickerFriendType: .talk,                //default .talk
                                                                    pickerFriendFilter: nil,                //default PickerFriendFilter.none
                                                                    viewAppearance: .auto,                   //default .auto
                                                                    countryCodes: nil,                      //default nil
                                                                    usingOsFilter: .all,                    //default .all
                                                                    orientation: .auto,                     //default .auto
                                                                    enableSearch: true,                     //default true
                                                                    enableIndex: true,                      //default true
                                                                    showMyProfile: true,                    //default true
                                                                    showFavorite: true,                     //default true
                                                                    disableSelectReasons: [.msgBlocked],    //defalut nil
                                                                    displayAllProfile: nil,                 //default false
                                                                    showPickedFriend: nil,                  //default true
                                                                    maxPickableCount: nil,                  //default 30
                                                                    minPickableCount: nil)                  //default 1
                    
                    PickerApi.shared.rx.selectFriends(params:pickerFriendRequestParams)
                        .subscribe (onNext:{ (selectedUsers) in
                            self.success(selectedUsers)

                            let _ = selectedUsers
                            //do something

                        }, onError: {error in
                            self.errorHandler(error: error)
                        })
                        .disposed(by: self.disposeBag)
                }),
                Row(title: "selectFriendsPopup(params:)", action: {
                    let pickerFriendRequestParams = PickerFriendRequestParams()
                    
                    PickerApi.shared.rx.selectFriendsPopup(params:pickerFriendRequestParams)
                        .subscribe (onNext:{ (selectedUsers) in
                            self.success(selectedUsers)

                            let _ = selectedUsers
                            //do something

                        }, onError: {error in
                            self.errorHandler(error: error)
                        })
                        .disposed(by: self.disposeBag)
                }),
                Row(title: "selectFriend(params:)", action: {
                    let pickerFriendRequestParams = PickerFriendRequestParams(title: "샘플앱 친구피커",                     //default "카카오톡 친구 선택"
                                                                    pickerFriendOrder: nil,                 //default nil
                                                                    pickerFriendType: .talk,                //default .talk
                                                                    pickerFriendFilter: nil,                //default PickerFriendFilter.none
                                                                    viewAppearance: .auto,                  //default .auto
                                                                    countryCodes: nil,                      //default nil
                                                                    usingOsFilter: .all,                    //default .all
                                                                    orientation: .auto,                     //default .auto
                                                                    enableSearch: true,                     //default true
                                                                    enableIndex: true,                      //default true
                                                                    showMyProfile: true,                    //default true
                                                                    showFavorite: true,                     //default true
                                                                    disableSelectReasons: [.msgBlocked],    //defalut nil
                                                                    displayAllProfile: nil,                 //default false
                                                                    showPickedFriend: nil,                  //ignore
                                                                    maxPickableCount: nil,                  //ignore
                                                                    minPickableCount: nil)                  //ignore
                    
                    PickerApi.shared.rx.selectFriend(params:pickerFriendRequestParams)
                        .subscribe (onNext:{ (selectedUsers) in
                            self.success(selectedUsers)

                            let _ = selectedUsers
                            //do something

                        }, onError: {error in
                            self.errorHandler(error: error)
                        })
                        .disposed(by: self.disposeBag)
                }),
                Row(title: "selectFriendPopup(params:)", action: {
                    let pickerFriendRequestParams = PickerFriendRequestParams()
                    
                    PickerApi.shared.rx.selectFriendPopup(params:pickerFriendRequestParams)
                        .subscribe (onNext:{ (selectedUsers) in
                            self.success(selectedUsers)

                            let _ = selectedUsers
                            //do something

                        }, onError: {error in
                            self.errorHandler(error: error)
                        })
                        .disposed(by: self.disposeBag)
                }),
            ]),
            // section FRIEND start =============================================================================================
            // MARK: - section auth start =============================================================================================
            SectionOfRow(header: "AUTH", items: [
                Row(title: "verifyAuthAge(authLevel:ageLimit:skipTerms:authFrom:adultsOnly:underAge:)", action: {
                    AuthApi.shared.rx.verifyAge(authLevel: .Level1,
                                                ageLimit: 25,
                                                skipTerms: false,
                                                adultsOnly: true,
                                                underAge: false)
                        .subscribe (onCompleted:{
                            self.success()
                        }, onError: { error in
                            self.errorHandler(error: error)
                        })
                        .disposed(by: self.disposeBag)
                })
            ]),
            // section auth end =============================================================================================
            
            // MARK: - section user start =============================================================================================
            SectionOfRow(header: "USER", items: [
                Row(title: "meForPartner()", action: {
                    UserApi.shared.rx.meForPartner()
                        .subscribe (onSuccess:{ (user) in
                            self.success()

                            let _ = user
                            //do something

                        }, onError: {error in
                            self.errorHandler(error: error)
                        })
                        .disposed(by: self.disposeBag)
                    
//                    UserApi.shared.rx.meForPartner(propertyKeys:["id"])
//                        .subscribe (onSuccess:{ (user) in
//                            self.success()
//
//                            let _ = user
//                            //do something
//
//                        }, onError: {error in
//                            self.errorHandler(error: error)
//                        })
//                        .disposed(by: self.disposeBag)
                }),
                Row(title: "signupForPartner()", action: {
                    UserApi.shared.rx.signupForPartner()
                        .subscribe (onSuccess:{ (userId) in
                            self.success()
                            
                            let _ = userId
                            //do something
                            
                        }, onError: {error in
                            self.errorHandler(error: error)
                        })
                        .disposed(by: self.disposeBag)
                }),
                Row(title: "ageAuthInfo()", action: {
                    UserApi.shared.rx.ageAuthInfo(ageLimit: 30)
                        .subscribe (onSuccess:{ (ageAuthInfo) in
                            self.success()
                            
                            let _ = ageAuthInfo
                            //do something
                            
                        }, onError: {error in
                            self.errorHandler(error: error)
                        })
                        .disposed(by: self.disposeBag)
                }),
                Row(title: "upgradeScopes(scopes:) - friends", action: {
                    UserApi.shared.rx.upgradeScopes(scopes:["friends"], guardianToken: nil)
                        .subscribe (onSuccess:{ (scopeInfo) in
                            self.success()
                            
                            let _ = scopeInfo
                            //do something
                            
                        }, onError: {error in
                            self.errorHandler(error: error)
                        })
                        .disposed(by: self.disposeBag)
                })
            ]),
            // section user end =============================================================================================
            
            // MARK: - section talk start =============================================================================================
            SectionOfRow(header: "TALK", items: [
                Row(title: "chatList()", action: {
                    TalkApi.shared.rx.chatList()
                        .retryWhen(Auth.shared.rx.incrementalAuthorizationRequired())
                        .subscribe (onSuccess:{ (chatList) in
                            self.success()
                            
                            let _ = chatList
                            //do something
                            
                        }, onError: {error in
                            self.errorHandler(error: error)
                        })
                        .disposed(by: self.disposeBag)
                }),
                Row(title: "chatMembers(chatId:friendsOnly:)", action: {
                    TalkApi.shared.rx.chatList()
                        .retryWhen(Auth.shared.rx.incrementalAuthorizationRequired())
                        .flatMap ({ (chatList) -> Single<ChatMembers> in
                            if let chats = chatList.elements?.first {
                                return TalkApi.shared.rx.chatMembers(chatId: chats.id, friendsOnly: false)
                                    .retryWhen(Auth.shared.rx.incrementalAuthorizationRequired())
                            }
                            else {
                                throw SdkError(reason:.Unknown, message: "chats empty!")
                            }
                        })
                        .subscribe (onSuccess:{ (chatMembers) in
                            self.success()
                            
                            let _ = chatMembers
                            //do something
                            
                        }, onError: {error in
                            self.errorHandler(error: error)
                        })
                        .disposed(by: self.disposeBag)
                }),
                Row(title: "friendsForPartner()", action: {
                    TalkApi.shared.rx.friendsForPartner()
                        .retryWhen(Auth.shared.rx.incrementalAuthorizationRequired())
                        .subscribe (onSuccess:{ (friends) in
                            self.success()
                            
                            let _ = friends
                            //do something
                            
                        }, onError: {error in
                            self.errorHandler(error: error)
                        })
                        .disposed(by: self.disposeBag)
                }),
                Row(title: "friendsForPartner(offset:limit:order:countryCodes:)", action: {
                    TalkApi.shared.rx.friendsForPartner(offset: 0, limit: 1, order: .Desc, countryCodes:["kr", "us"])
                        .retryWhen(Auth.shared.rx.incrementalAuthorizationRequired())
                        .subscribe (onSuccess:{ (friends) in
                            self.success()
                            
                            let _ = friends
                            //do something
                            
                        }, onError: {error in
                            self.errorHandler(error: error)
                        })
                        .disposed(by: self.disposeBag)
                }),
                Row(title: "friendsForPartner(offset:limit:order:)", action: {
//                    Single.zip(
//                        TalkApi.shared.rx.friendsForPartner(offset: 0, limit: 1, order: .Desc).retryWhen(Auth.shared.rx.incrementalAuthorizationRequired()),
//                        TalkApi.shared.rx.friendsForPartner(offset: 1, limit: 1, order: .Desc).retryWhen(Auth.shared.rx.incrementalAuthorizationRequired()),
//                        TalkApi.shared.rx.friendsForPartner(offset: 2, limit: 1, order: .Desc).retryWhen(Auth.shared.rx.incrementalAuthorizationRequired()),
//                        resultSelector: { first, second, third in
//                            print(" ---->first:\(first)\n ---->second:\(second)\n ---->third:\(third)")
//                            return Friends<PartnerFriend>.merge([first, second, third])
//                    })
//                    .subscribe (onSuccess:{
//                        self.success($0)
//                    }, onError: {error in
//                        self.errorHandler(error: error)
//                    })
//                    .disposed(by: self.disposeBag)
                }),
                Row(title: "sendDefaultMessageForPartner(templatable:receiverUuids:)", action: {
                    let title = "카카오톡 링크 4.0(메세지)"
                    let description = "친구에게보내기 티폴트 템플릿 FEED"
                    
                    let feedTemplateJsonStringData =
                        """
                            {
                            "object_type": "feed",
                            "content": {
                            "title": "\(title)",
                            "description": "\(description)",
                            "image_url": "http://alpha-api1-kage.kakao.com/dn/cerDB5/ZSb2iRugKx/M4nuZxX823tnK1Mk5yVcv0/kakaolink40_original.png",
                            "link": {
                            "web_url": "http://dev.kakao.com",
                            "mobile_web_url": "http://dev.kakao.com"
                            }
                            },
                            "social": {
                            "like_count": 100,
                            "comment_count": 200
                            },
                            "button_title": "바로 확인"
                            }
                            """.data(using: .utf8)!
                    
                    if let templatable = try? SdkJSONDecoder.custom.decode(FeedTemplate.self, from: feedTemplateJsonStringData) {
                        TalkApi.shared.rx.friendsForPartner()
                            .retryWhen(Auth.shared.rx.incrementalAuthorizationRequired())
                            .subscribe(onSuccess: { (friends) in
                                self.success(friends, updateIndicator: false)
                                
                                if let elements = self.extractPickerElements(friends: friends) {
                                    let vc = RxSDKSamplePickerViewController(title: "Friends",
                                                                             elements: elements,
                                                                             elementsPickerCompletionHandler: { (selectedIds) in
                                                                                let friendUuids = selectedIds as! [String]
                                                                                    TalkApi.shared.rx.sendDefaultMessageForPartner(templatable:templatable, receiverUuids:friendUuids)
                                                                                        .retryWhen(Auth.shared.rx.incrementalAuthorizationRequired())
                                                                                        .subscribe(onSuccess:{ (partnerMessageSendResult) in
                                                                                            self.success()
                                                                                            
                                                                                            //do something
                                                                                            let _ = partnerMessageSendResult
                                                                                            

                                                                                        }, onError: {error in
                                                                                            self.errorHandler(error: error)
                                                                                        })
                                                                                        .disposed(by: self.disposeBag)
                                                                                
                                    })
                                    self.navigationController?.pushViewController(vc, animated: true)
                                }
                            }, onError: {error in
                                self.errorHandler(error: error)
                            })
                            .disposed(by: self.disposeBag)
                    }
                }),
                Row(title: "sendDefaultMessageForPartner(templatable:receiverChatIds:)", action: {
                    let title = "카카오톡 링크 4.0(메세지)"
                    let description = "친구에게보내기 티폴트 템플릿 FEED"
                    
                    let feedTemplateJsonStringData =
                        """
                            {
                            "object_type": "feed",
                            "content": {
                            "title": "\(title)",
                            "description": "\(description)",
                            "image_url": "http://alpha-api1-kage.kakao.com/dn/cerDB5/ZSb2iRugKx/M4nuZxX823tnK1Mk5yVcv0/kakaolink40_original.png",
                            "link": {
                            "web_url": "http://dev.kakao.com",
                            "mobile_web_url": "http://dev.kakao.com"
                            }
                            },
                            "social": {
                            "like_count": 100,
                            "comment_count": 200
                            },
                            "button_title": "바로 확인"
                            }
                            """.data(using: .utf8)!
                    
                    if let templatable = try? SdkJSONDecoder.custom.decode(FeedTemplate.self, from: feedTemplateJsonStringData) {
                        TalkApi.shared.rx.chatList()
                            .retryWhen(Auth.shared.rx.incrementalAuthorizationRequired())
                            .subscribe(onSuccess: { (chats) in
                                self.success(chats, updateIndicator: false)
                                
                                if let elements = self.extractPickerElements(chats: chats) {
                                    let vc = RxSDKSamplePickerViewController(title: "Chats",
                                                                             elements: elements,
                                                                             elementsPickerCompletionHandler: { (selectedIds) in
                                                                                let chatIds = selectedIds as! [Int64]
                                                                                TalkApi.shared.rx.sendDefaultMessageForPartner(templatable: templatable, receiverChatIds: chatIds)
                                                                                    .retryWhen(Auth.shared.rx.incrementalAuthorizationRequired())
                                                                                    .subscribe(onSuccess:{ (partnerMessageSendResult) in
                                                                                        self.success()
                                                                                        
                                                                                        //do something
                                                                                        let _ = partnerMessageSendResult
                                                                                        

                                                                                    }, onError: {error in
                                                                                        self.errorHandler(error: error)
                                                                                    })
                                                                                    .disposed(by: self.disposeBag)
                                    })
                                    self.navigationController?.pushViewController(vc, animated: true)
                                }
                            }, onError: {error in
                                self.errorHandler(error: error)
                            })
                            .disposed(by: self.disposeBag)
                    }
                }),
                Row(title: "sendCustomMessageForPartner(templateId:receiverUuids:)", action: {
                    TalkApi.shared.rx.friendsForPartner()
                        .retryWhen(Auth.shared.rx.incrementalAuthorizationRequired())
                        .subscribe(onSuccess: { (friends) in
                            self.success(friends, updateIndicator: false)
                            
                            if let elements = self.extractPickerElements(friends: friends) {
                                let vc = RxSDKSamplePickerViewController(title: "Friends",
                                                                         elements: elements,
                                                                         elementsPickerCompletionHandler: { (selectedIds) in
                                                                            let friendUuids = selectedIds as! [String]
                                                                            TalkApi.shared.rx.sendCustomMessageForPartner(templateId:637, receiverUuids:friendUuids)
                                                                                .retryWhen(Auth.shared.rx.incrementalAuthorizationRequired())
                                                                                .subscribe(onSuccess:{ (partnerMessageSendResult) in
                                                                                    self.success()
                                                                                    
                                                                                    //do something
                                                                                    let _ = partnerMessageSendResult
                                                                                    

                                                                                }, onError: {error in
                                                                                    self.errorHandler(error: error)
                                                                                })
                                                                                .disposed(by: self.disposeBag)
                                                                            
                                })
                                self.navigationController?.pushViewController(vc, animated: true)
                            }
                        }, onError: {error in
                            self.errorHandler(error: error)
                        })
                        .disposed(by: self.disposeBag)
                }),
                Row(title: "sendCustomMessageForPartner(templateId:receiverChatIds:)", action: {
                    TalkApi.shared.rx.chatList()
                        .retryWhen(Auth.shared.rx.incrementalAuthorizationRequired())
                        .subscribe(onSuccess: { (chats) in
                            self.success(chats, updateIndicator: false)
                            
                            if let elements = self.extractPickerElements(chats: chats) {
                                let vc = RxSDKSamplePickerViewController(title: "Chats",
                                                                         elements: elements,
                                                                         elementsPickerCompletionHandler: { (selectedIds) in
                                                                            let chatIds = selectedIds as! [Int64]
                                                                            TalkApi.shared.rx.sendCustomMessageForPartner(templateId:637, receiverChatIds:chatIds)
                                                                                .retryWhen(Auth.shared.rx.incrementalAuthorizationRequired())
                                                                                .subscribe(onSuccess:{ (partnerMessageSendResult) in
                                                                                    self.success()
                                                                                    
                                                                                    //do something
                                                                                    let _ = partnerMessageSendResult
                                                                                    

                                                                                }, onError: {error in
                                                                                    self.errorHandler(error: error)
                                                                                })
                                                                                .disposed(by: self.disposeBag)
                                                                            
                                })
                                self.navigationController?.pushViewController(vc, animated: true)
                            }
                        }, onError: {error in
                            self.errorHandler(error: error)
                        })
                        .disposed(by: self.disposeBag)
                }),
                Row(title: "sendCustomMessage(templateId:templateArgs:receiverUuids:)", action: {
                    TalkApi.shared.rx.friendsForPartner()
                        .retryWhen(Auth.shared.rx.incrementalAuthorizationRequired())
                        .subscribe(onSuccess: { (friends) in
                            self.success(friends, updateIndicator: false)
                            
                            if let elements = self.extractPickerElements(friends: friends) {
                                let vc = RxSDKSamplePickerViewController(title: "Friends",
                                                                         elements: elements,
                                                                         elementsPickerCompletionHandler: { (selectedIds) in
                                                                            let friendUuids = selectedIds as! [String]
                                                                            TalkApi.shared.rx.sendCustomMessageForPartner(templateId:664, templateArgs:["title":"제목입니다.", "description":"설명입니다."], receiverUuids:friendUuids)
                                                                                .retryWhen(Auth.shared.rx.incrementalAuthorizationRequired())
                                                                                .subscribe(onSuccess:{ (partnerMessageSendResult) in
                                                                                    self.success()
                                                                                    
                                                                                    //do something
                                                                                    let _ = partnerMessageSendResult
                                                                                    

                                                                                }, onError: {error in
                                                                                    self.errorHandler(error: error)
                                                                                })
                                                                                .disposed(by: self.disposeBag)
                                                                            
                                })
                                self.navigationController?.pushViewController(vc, animated: true)
                            }
                        }, onError: {error in
                            self.errorHandler(error: error)
                        })
                        .disposed(by: self.disposeBag)

                }),
                Row(title: "sendCustomMessageForPartner(templateId:templateArgs:receiverChatIds:)", action: {
                    TalkApi.shared.rx.chatList()
                        .retryWhen(Auth.shared.rx.incrementalAuthorizationRequired())
                        .subscribe(onSuccess: { (chats) in
                            self.success(chats, updateIndicator: false)
                            
                            if let elements = self.extractPickerElements(chats: chats) {
                                let vc = RxSDKSamplePickerViewController(title: "Chats",
                                                                         elements: elements,
                                                                         elementsPickerCompletionHandler: { (selectedIds) in
                                                                            let chatIds = selectedIds as! [Int64]
                                                                            TalkApi.shared.rx.sendCustomMessageForPartner(templateId:664, templateArgs:["title":"제목입니다.", "description":"설명입니다."], receiverChatIds:chatIds)
                                                                                .retryWhen(Auth.shared.rx.incrementalAuthorizationRequired())
                                                                                .subscribe(onSuccess:{ (partnerMessageSendResult) in
                                                                                    self.success()
                                                                                    
                                                                                    //do something
                                                                                    let _ = partnerMessageSendResult
                                                                                    

                                                                                }, onError: {error in
                                                                                    self.errorHandler(error: error)
                                                                                })
                                                                                .disposed(by: self.disposeBag)
                                                                            
                                })
                                self.navigationController?.pushViewController(vc, animated: true)
                            }
                        }, onError: {error in
                            self.errorHandler(error: error)
                        })
                        .disposed(by: self.disposeBag)
                }),
                Row(title: "sendScrapMessageForPartner(requestUrl:receiverUuids:)", action: {
                    TalkApi.shared.rx.friendsForPartner()
                        .retryWhen(Auth.shared.rx.incrementalAuthorizationRequired())
                        .subscribe(onSuccess: { (friends) in
                            self.success(friends, updateIndicator: false)
                            
                            if let elements = self.extractPickerElements(friends: friends) {
                                let vc = RxSDKSamplePickerViewController(title: "Friends",
                                                                         elements: elements,
                                                                         elementsPickerCompletionHandler: { (selectedIds) in
                                                                            let friendUuids = selectedIds as! [String]
                                                                            TalkApi.shared.rx.sendScrapMessageForPartner(requestUrl:"https://developers.kakao.com", receiverUuids:friendUuids)
                                                                                .retryWhen(Auth.shared.rx.incrementalAuthorizationRequired())
                                                                                .subscribe(onSuccess:{ (partnerMessageSendResult) in
                                                                                    self.success()
                                                                                    
                                                                                    //do something
                                                                                    let _ = partnerMessageSendResult
                                                                                    

                                                                                }, onError: {error in
                                                                                    self.errorHandler(error: error)
                                                                                })
                                                                                .disposed(by: self.disposeBag)
                                                                            
                                })
                                self.navigationController?.pushViewController(vc, animated: true)
                            }
                        }, onError: {error in
                            self.errorHandler(error: error)
                        })
                        .disposed(by: self.disposeBag)
                }),
                Row(title: "sendScrapMessageForPartner(requestUrl:templateId:receiverChatIds:)", action: {
                    TalkApi.shared.rx.chatList()
                        .retryWhen(Auth.shared.rx.incrementalAuthorizationRequired())
                        .subscribe(onSuccess: { (chats) in
                            self.success(chats, updateIndicator: false)
                            
                            if let elements = self.extractPickerElements(chats: chats) {
                                let vc = RxSDKSamplePickerViewController(title: "Chats",
                                                                         elements: elements,
                                                                         elementsPickerCompletionHandler: { (selectedIds) in
                                                                            let chatIds = selectedIds as! [Int64]
                                                                            TalkApi.shared.rx.sendScrapMessageForPartner(requestUrl: "https://developers.kakao.com", templateId:665, receiverChatIds:chatIds)
                                                                                .retryWhen(Auth.shared.rx.incrementalAuthorizationRequired())
                                                                                .subscribe(onSuccess:{ (partnerMessageSendResult) in
                                                                                    self.success()
                                                                                    
                                                                                    //do something
                                                                                    let _ = partnerMessageSendResult
                                                                                    

                                                                                }, onError: {error in
                                                                                    self.errorHandler(error: error)
                                                                                })
                                                                                .disposed(by: self.disposeBag)
                                                                            
                                })
                                self.navigationController?.pushViewController(vc, animated: true)
                            }
                        }, onError: {error in
                            self.errorHandler(error: error)
                        })
                        .disposed(by: self.disposeBag)
                }),
                
            ]),
            // section talk end =============================================================================================
            
            // MARK: - section etc start =============================================================================================
            SectionOfRow(header: "ETC", items: [
                Row(title: "token() usage", action: {
                    let disposeBag = DisposeBag()
                    
                    AuthApi.shared.rx.token(code: "#발급받은코드#")
                        .subscribe (onSuccess:{ (user) in
                            //success
                            //토큰은 SDK 내부 저장소에 저장됩니다.(이후 SDK사용은 기존과 똑같이 하면됩니다.)
                            
                        }, onError: {error in
                            print(error)
                        })
                        .disposed(by: disposeBag)
                    
                }),
                Row(title: "crypto test", action: {
                    
                    let accessToken = "QPnCpaUplT6z_w2tL0KdudHJw4mxSQpg1rlzvArCHesAAAFuz29kRA"
                    let keyString = "SDK-417E6CB3-BC79-48BC-82DD-D4603BAA426E"
                                        
                    print(">accessToken : \(accessToken)")
                    
                    let accessTokenData = accessToken.data(using: .utf8)
                    print(">accessTokenData : \(String(describing:accessTokenData))")
                    print(String(describing:accessTokenData?.hexEncodedString()))
                    
                    print(">keyString: \(keyString)")
                    let cryptedTokenData = SdkCrypto.shared.encryptForMigration(data: accessTokenData, key:keyString.data(using: .utf8))
                    print(">cryptedTokenData: \(String(describing:cryptedTokenData))")
                    print(String(describing:cryptedTokenData?.hexEncodedString()))
                    
                    if let decryptedTokenData = SdkCrypto.shared.decryptForMigration(data: cryptedTokenData, key:keyString.data(using: .utf8)) {
                        print(">decryptedTokenData: \(decryptedTokenData)")
                        print(decryptedTokenData.hexEncodedString())
                        print(">decryptedToken: \(String(describing:String(bytes: decryptedTokenData, encoding: .utf8)))")
                    }
                    else {
                        print(">decryptedToken: nil")
                    }
                }),
                Row(title: "loginWithKakaoAccountAtAccountPage(accountParameters:)-[view:login],[email:a@kakao.com]", action: {
                    UserApi.shared.rx.loginWithKakaoAccount(accountParameters: ["view":"login", "email":"a@kakao.com"])
                        .subscribe(onNext:{ (oauthToken) in
                            self.success()
                            
                            //do something
                            _ = oauthToken
                            
                        }, onError: {error in
                            self.errorHandler(error: error)
                        })
                        .disposed(by: self.disposeBag)
                }),
                Row(title: "loginWithKakaoAccountAtAccountPage(accountParameters:)-[view:signup]", action: {
                    UserApi.shared.rx.loginWithKakaoAccount(accountParameters: ["view":"signup"])
                        .subscribe (onNext:{ (user) in
                            self.success()
                            //do something
                            
                        }, onError: {error in
                            self.errorHandler(error: error)
                        })
                        .disposed(by: self.disposeBag)
                }),
                Row(title: "loginWithKakaoAccountAtAccountPage(accountParameters:)-[:]", action: {
                    UserApi.shared.rx.loginWithKakaoAccount(accountParameters: [:])
                        .subscribe (onNext:{ (user) in
                            self.success()
                            //do something
                            
                        }, onError: {error in
                            self.errorHandler(error: error)
                        })
                        .disposed(by: self.disposeBag)
                })
            ]),
            // section etc start =============================================================================================
        ]
    }
}
